(* compiler.ml *)
(* A compiler for simple arithmetic expressions. *)

(******************************************************************************)

open Printf
open Ast
open X86   (* Note that Ast has similarly named constructors that must be 
              disambiguated.  For example: Ast.Shl vs. X86.Shl *)

(* A stream is just a list of instructions, with the invariant that they will
 * be executed in *reverse* list order. *)
type stream = insn list

(* Some notation to make reversed list consing more readable.                *)
(* str >:: insn   means "run the instructions in stream and then do insn" *)
let (>::) x y = y::x


(* Builds a stream (reversed list) of instructions for the given AST. The second
 * argument is an input stream that will be extended with the code emitted
 * for this expression. *)
let rec emit_exp (e:exp) (str : stream) : stream = 
  let tmploc = stack_offset 0l in
  begin match e with
    | Cint i -> str >:: (Mov (eax, Imm i))
    | Arg    -> str >:: (Mov (eax, edx))
    | Plus(l,r) ->
	let str_l = 
	  (emit_exp l str) >:: (* Compute the left-hand expression into Eax *)
          (Push eax)                (* Store it to the stack *)
	in             
	(emit_exp r str_l) >::  (* Compute the right-hand exp (after the left) *)
	(Add (eax, tmploc)) >:: (* Do the addition. *)
	(Add (esp, Imm 4l))     (* Discard the stack-saved value *)
  end

(* Builds a globally-visible X86 instruction block that acts like the C fuction:

   int program(int X) { return <expression>; }

   Follows cdecl calling conventions and platform-specific name mangling policy. *)
let compile_exp (ast:exp) : Cunit.cunit =
  let block_name = (Platform.decorate_cdecl "program") in
  let init_str = [Mov(edx, stack_offset(4l))] in
  let insns = List.rev ((emit_exp ast init_str) >:: Ret) in
  let main_bk = {
    global = true;
    label  = mk_lbl_named block_name;
    insns = insns;
  } in [Cunit.Code main_bk]

