open Ast

(* Simple let language for the arithmetic language *)

(* Unique identifiers for temporaries. *)
type uid = int

(* "gensym" -- generate a new unique identifier *)
let mk_uid : unit -> uid =
  let ctr = ref 0 in
    fun () -> let uid = !ctr in
	ctr := !ctr + 1;
	uid

(* The type of operands *)
type opnd = 
  | Tmp of uid
  | Imm of int32

(* The type of operations *)
(* Note that there is no nesting of operations! *)
type op =
  | GetArg                (* pseudo instruction to get the value of X *)
  | Add of opnd * opnd    

(* The type of commands *)
type cmd =
  | Let of uid * op * cmd
  | Ret of opnd

(* Translation Strategy 1:
 *   convert the expressions into a stream of 
 *   tuples of the form "uid * op" where uid is the 
 *   identifier to be let-bound to the given op.
 *
 *  Invariant: emit_exp returns the "answer" together
 *  with the stream.
 *
 *  Afterwards, post-process this stream to turn it
 *  into a sequence of nested commands.
 *)
type stream = (uid * op) list
let (>::) x y = y::x   (* snoc (a.k.a. reverse cons) *)

let rec emit_exp (e:exp) (str:stream) : opnd * stream =
  begin match e with
    | Cint i -> (Imm i, str)
    | Arg    -> 
	let tmp = mk_uid () in
	  (Tmp tmp, str >:: (tmp, GetArg))
    | Plus(l, r) ->
	let (ansl, str_l) = emit_exp l str in
	let (ansr, str_r) = emit_exp r str_l in
	let tmp = mk_uid () in
	  (Tmp tmp, str_r >:: (tmp, Add(ansl, ansr)))
  end

let sll_exp (e:exp) : cmd =
  let (ret, str) = emit_exp e [] in
    List.fold_left (fun (c:cmd) -> fun (v, o) -> Let(v, o, c)) (Ret ret) str
    


(* Translation strategy 2:
 *   "CPS"-style translation.
 *
 *  Use a "meta-level" function to say "what to do next".
 *  This meta-level function returns object-level code.
 *)

let rec sll_cps (e:exp) (k:opnd -> cmd) : cmd =
  begin match e with
    | Cint i -> k (Imm i)
    | Arg    -> 
	let tmp = mk_uid () in
	  Let (tmp, GetArg, k (Tmp tmp))
    | Plus(l, r) ->
	sll_cps l 
	  (fun ansl -> sll_cps r 
	     (fun ansr ->
		let tmp = mk_uid () in
		  Let (tmp, Add(ansl, ansr), k (Tmp tmp))))
  end
			  
let rec sll_exp_cps (e:exp) : cmd =
    sll_cps e (fun ans -> Ret ans)


(* Interpretting via CPS
 * 
 * The meta-level function returns meta-level values.
 * Note the type annotations.
 *
 * Since the top-level program contains the parameter
 * X, we have to add an extra argument to the function.
 *)
let rec sll_interp (x:int32) (e:exp) (k:int32 -> int32) : int32 =
  begin match e with
    | Cint i -> k i
    | Arg -> k x    (* Here we just "return" the argument x *)
    | Plus(l, r) ->
	sll_interp x l 
	  (fun ansl -> sll_interp x r
	     (fun ansr ->
		k (Int32.add ansl ansr)))
  end
	
(* The initial continuation is just the identity function *)
let interp (x:int32) (e:exp) : int32 =
  sll_interp x e (fun x -> x)


(*** Printing **)
let opnd_to_string o =
  begin match o with
    | Tmp x -> Printf.sprintf "tmp%d" x
    | Imm l -> Int32.to_string l
  end

let op_to_string o =
  begin match o with
    | GetArg -> "GetArg"
    | Add(o1, o2) -> Printf.sprintf "Add(%s, %s)" 
	(opnd_to_string o1) (opnd_to_string o2)
  end

let rec cmd_to_string c =
  begin match c with
    | Let(x, o, body) ->
      let cstr = cmd_to_string body in
	Printf.sprintf "let %s = %s in\n%s" 
	  (opnd_to_string (Tmp x))
	  (op_to_string o)
	  cstr
    | Ret o -> Printf.sprintf "  return %s" (opnd_to_string o)
  end


